using System;
using System.Collections.Generic;
using Core.Actions;
using Core.Domain;
using UniRx;
using Utils;

namespace Presentation
{
    public class HeadPresenter
    {
        private readonly HeadView view;
        private readonly RetrieveHead retrieveHead;
        private readonly TouchBrainSector touchBrainSector;
        private readonly BotAction botAction;
        private readonly GetCurrentPoints getCurrentPoints;

        private Disposer disposer = new Disposer();

        private const float BRAIN_SECTOR_COOLDOWN = 0.1f;
        
        Dictionary<BrainSector, DateTime> brainSectorsCooldownsMap = new Dictionary<BrainSector, DateTime>();
        
        public HeadPresenter(HeadView view,
            RetrieveHead retrieveHead,
            TouchBrainSector touchBrainSector, 
            CreateHead createHead,
            BotAction botAction,
            GetCurrentPoints getCurrentPoints)
        {
            createHead.Execute();
            
            this.view = view;
            this.retrieveHead = retrieveHead;
            this.touchBrainSector = touchBrainSector;
            this.botAction = botAction;
            this.getCurrentPoints = getCurrentPoints;

            view.OnBrainSectorTouched += TouchBrainSector;
            view.OnStart += Present;
            view.OnExit += Clean;
            
            view.OnEveryUpdate()
                .Do(_ => DoBotAction())
                .Do(_ => UpdatePoints())
                .Subscribe()
                .AddTo(disposer);
        }

        private void UpdatePoints()
        {
            view.UpdateCurrentPoints(getCurrentPoints.Execute());
        }

        private void DoBotAction()
        {
            botAction.Execute();
            var updatedHead = retrieveHead.Execute();
            UpdateHeadView(updatedHead);
        }

        public void Present()
        {
            view.ShowHead(retrieveHead.Execute().GetHeadPieces());
        }

        private void TouchBrainSector(BrainSector brainSector)
        {
            if (IsBrainSectorOnCooldown(brainSector))
                return;

            view.BrainTouchedSfx();
            AddBrainSectorToCooldown(brainSector);
            UpdateHeadView(touchBrainSector.Execute(brainSector));
        }

        private void AddBrainSectorToCooldown(BrainSector brainSector)
        {
            view.ShowBrainSectorCooldown(brainSector, BRAIN_SECTOR_COOLDOWN);
            brainSectorsCooldownsMap[brainSector] = DateTime.Now.AddSeconds(BRAIN_SECTOR_COOLDOWN);
        }

        private bool IsBrainSectorOnCooldown(BrainSector brainSector)
        {
            if (brainSectorsCooldownsMap.ContainsKey(brainSector))
                return brainSectorsCooldownsMap[brainSector] >= DateTime.Now;

            return false;
        }

        private void CheckIfGameIsOver()
        {
            if (retrieveHead.Execute().IsDead())
                EndGame();
        }

        private void UpdateHeadView(Head updatedHead)
        {
            view.ShowHead(updatedHead.GetHeadPieces());
            
            CheckIfGameIsOver();
        }

        private void EndGame()
        {
            Clean();
            view.OnBrainSectorTouched -= TouchBrainSector;
            view.FinishGame(getCurrentPoints.Execute());
        }

        private void Clean()
        {
            disposer.Dispose();
        }
    }
}