using System;
using System.Collections;
using System.Collections.Generic;
using Core.Domain;
using Infrastructure.Providers;
using Presentation;
using TMPro;
using UniRx;
using UnityEngine;
using UnityEngine.UI;

namespace UnityDelivery
{
    public class UnityHeadView : MonoBehaviour, HeadView
    {
        public event Action<BrainSector> OnBrainSectorTouched;
        public event Action OnStart;
        public event Action OnExit;

        [SerializeField] private GameObject leftEye;
        [SerializeField] private GameObject rightEye;
        [SerializeField] private GameObject nose;
        [SerializeField] private GameObject leftMouth;
        [SerializeField] private GameObject rightMouth;
        [SerializeField] private GameObject foreHead;
        [SerializeField] private GameObject chin;
        
        [SerializeField] private List<BrainSectorView> brainSectorViews;

        [SerializeField] private TextMeshProUGUI currentPointsText;

        [SerializeField] private AudioSource clickSfx;
        [SerializeField] private AudioSource inGameMusic;
        [SerializeField] private AudioSource endGameMusic;
        
        
        readonly Dictionary<BrainSector, BrainSectorView> brainSectorViewsMap = new Dictionary<BrainSector, BrainSectorView>();
        readonly Dictionary<HeadPieceType, GameObject> headPiecesObjects = new Dictionary<HeadPieceType, GameObject>();


        [SerializeField] private EndScreen finishEndScreen;
        public void Awake()
        {
            SetHeadPiecesDictionary();
            SetUpBrainSectorsActions();
            
            HeadPresenterProvider.For(this);
        }

        public void Start()
        {
            OnStart();
        }

        public void OnDisable()
        {
            OnExit();
        }

        public void OnDestroy()
        {
            OnExit();
        }

        private void SetUpBrainSectorsActions()
        {
            foreach (var brainSectorView in brainSectorViews)
            {
                brainSectorViewsMap.Add(brainSectorView.BrainSector, brainSectorView);
                brainSectorView.button.onClick.AddListener(() => OnBrainSectorTouched(brainSectorView.BrainSector));
            }
        }
        
        private void SetHeadPiecesDictionary()
        {
            headPiecesObjects.Add(HeadPieceType.LeftEye, leftEye);
            headPiecesObjects.Add(HeadPieceType.RightEye, rightEye);
            headPiecesObjects.Add(HeadPieceType.Nose, nose);
            headPiecesObjects.Add(HeadPieceType.LeftMouth, leftMouth);
            headPiecesObjects.Add(HeadPieceType.RightMouth, rightMouth);
            headPiecesObjects.Add(HeadPieceType.ForeHead, foreHead);
            headPiecesObjects.Add(HeadPieceType.Chin, chin);
        }

        public void ShowHead(List<HeadPiece> headPieces)
        {
            foreach (var headPiece in headPieces)
            {
                if(headPiecesObjects.ContainsKey(headPiece.GetHeadType()))
                    headPiecesObjects[headPiece.GetHeadType()].SetActive(headPiece.IsBroken);
            }
        }

        public IObservable<Unit> OnEveryUpdate()
        {
            return Observable.EveryUpdate()
                .AsUnitObservable();
        }

        public void FinishGame(double score)
        {
            inGameMusic.Stop();
            endGameMusic.Play();
            
            StartCoroutine(ShowScorePopUp(score));
        }

        
        private IEnumerator ShowScorePopUp(double score)
        {
            yield return new WaitForSeconds(2f);
            
            finishEndScreen.gameObject.SetActive(true);
            finishEndScreen.ShowScore(score);
        }

        public void ShowBrainSectorCooldown(BrainSector brainSector, float brainSectorCooldown)
        {
            if(brainSectorViewsMap.ContainsKey(brainSector))
                brainSectorViewsMap[brainSector].animator.SetTrigger("Cooldown");
        }

        public void UpdateCurrentPoints(double currentPoints)
        {
            currentPointsText.SetText(String.Format("{0:C0}", currentPoints));
        }

        public void BrainTouchedSfx()
        {
            clickSfx.Play();
        }
    }

    [Serializable]
    public class BrainSectorView
    {
        [SerializeField] public BrainSector BrainSector;
        [SerializeField] public Button button;
        [SerializeField] public Animator animator;
    }
}